(ns puppetlabs.rbac.services.http.api-validation
  (:require [clojure.java.io :as io]
            [cheshire.core :as json]
            [schema.core :as sc]
            [slingshot.slingshot :refer [throw+]]
            [puppetlabs.rbac.schema :as rs]
            [puppetlabs.kitchensink.core :as ks]
            [puppetlabs.schema-tools :refer [explain-and-simplify-exception-data]])
  (:import com.fasterxml.jackson.core.JsonParseException))

(defn body-as-string
  "Given a Liberator context, convert its request body into a string
   and return it."
  [ctx]
  (when-let [body (get-in ctx [:request :body])]
    (condp instance? body
      java.lang.String body
      (slurp (io/reader body)))))

(defn extract-body
  "Given a Liberator context with a JSON request body, deserialize it
   and make it ready to be used in Clojure."
  [ctx]
  (-> ctx
      body-as-string
      (json/parse-string true)))

(defn valid-id-filter? [id-str]
  (some? (re-matches #"(\d,?)+" id-str)))

(defn id-filter->vec [id-str]
  (mapv ks/parse-int
        (clojure.string/split id-str #",")))

(defn extract-id-filter
  "Given a query parameter representing a list of strings, validate it
  as a list of integers and then convert it into a vector. Returns a
  form suitable for :malformed?"
  [id-str]
  (cond
   (nil? id-str) false
   (valid-id-filter? id-str) [false {:id-filter (id-filter->vec id-str)}]
   :else (throw+ {:kind :puppetlabs.rbac/invalid-id-filter
                  :error "Invalid ID filter list"})))

(sc/defn validate-id-key!
  "Takes the id from the route and a body. If the body contains an :id key,
   will throw an error if the value of :id does not match the id provided."
  [id :- (sc/maybe sc/Int), body]
  (cond
    (nil? id)         nil
    (nil? (:id body)) nil
    (= (:id body) id) nil
    :else (throw+ {:kind :puppetlabs.rbac/inconsistent-id
                   :error "ID in request must match route"
                   :url-id id
                   :body-id (:id body)})))

(defn validate-schema!
  "This wraps schema/check and throws a custom slingshot stone instead of the
  generic exception that schema/validate would throw."
  [schema body]
  (when-let [schema-error (sc/check schema body)]
    (throw+ {:kind :puppetlabs.rbac/schema-violation
             :schema schema
             :submitted body
             :error schema-error})))

(sc/defn liberator-validate!
  "This function returns output intended for Liberator's :malformed?
   decision point. Takes a schema, a keyword, and the body to validate.
   If validation is successful, it will return the following:

       [false, {:<key> <body>}]

   This response indicates to Liberator that the response is NOT malformed
   ('false') and will associate the body to the context under the key
   provided.

   If validation is not successful, it will throw a slingshot stone with :kind
   set to :puppetlabs.rbac/schema-violation."

  [schema :- sc/Schema, key :- sc/Keyword, body & [id]]
  (validate-id-key! id body)
  (validate-schema! schema body)
  [false {key body}])

(defn- err-with-rep
  "Necessary when returning errors early in liberator's decision chain (e.g.
  from `malformed?`) because it doesn't know how to render error responses
  until the `media-type-available?` decision has been executed"
  [err]
  (assoc err :representation {:media-type "application/json"}))

(sc/defn extract-and-validate
  "Validates request body against the schema, and returns the result in
   the appropriate form for Liberator."
  [schema :- sc/Schema, key :- sc/Keyword, ctx & id]
  (try
    (let [body (extract-body ctx)]
      (apply liberator-validate! schema key body id))
    (catch JsonParseException e
      [true (err-with-rep {:error e})])))

(defn directory-settings-validate
  "Adds special handling for an empty body, which signals a directory
   server disconnect and is always considered non-malformed. Otherwise
   delegates to liberator-validate!."
  [key ctx]
  (let [body (extract-body ctx)]
    (if (empty? body)
      false
      (liberator-validate! rs/DirectoryServiceSettings key body))))

(defn handle-malformed
  [ctx]
  {:kind "malformed-request"
   :msg "The body of your request is not valid JSON."
   :details {:error (if (:error ctx)
                      (.toString (:error ctx))
                      nil)}})
